/*
 TUIO C# Library - part of the reacTIVision project
 Copyright (c) 2005-2014 Martin Kaltenbrunner <martin@tuio.org>

 This library is free software; you can redistribute it and/or
 modify it under the terms of the GNU Lesser General Public
 License as published by the Free Software Foundation; either
 version 3.0 of the License, or (at your option) any later version.
 
 This library is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 Lesser General Public License for more details.
 
 You should have received a copy of the GNU Lesser General Public
 License along with this library.
*/
/* WallToken extension by Anonymous authors (2020) **/

using System;
using System.Collections.Generic;
using TUIO;

namespace WTT
{
    /**
     * <remarks>
     * The TuioWallToken class encapsulates either a /tuio/2DToken or a TUIO objects.
     * </remarks>
     *
     * @author Anonymous author
     * @version 1.1.5
     */
    
    public class TuioWallToken : TuioObject
    {

        /**
        * <summary>
        * The individual name ID that is assigned to each TuioCursor in the template.</summary>
        */
        protected string tokenID;

        /**
        * <summary>
        * tell whether the token is fixed on the wall.</summary>
        */
        protected bool tfixed;

        /**
        * <summary>
        * Color of the token </summary>
        */
        protected int[] rgba = new int[4];

        protected string shape = "";
        protected float ray = 0;
        protected float inner = 0;
        
        #region Constructors

        /**
         * <summary>
         * This constructor takes a TuioTime argument and assigns it along with the provided
         * Session ID, Symbol ID, X and Y coordinate and angle to the newly created TuioObject.</summary>
         *
         * <param name="ttime">the TuioTime to assign</param>
         * <param name="si">the Session ID to assign</param>
         * <param name="sym">the Symbol ID to assign</param>
         * <param name="tid">the name ID to assign</param>
         * <param name="xp">the X coordinate to assign</param>
         * <param name="yp">the Y coordinate to assign</param>
         * <param name="a">the angle to assign</param>
         */
        public TuioWallToken(TuioTime ttime, long si, int sym, string tid, float xp, float yp, float a)
            : base(ttime, si, sym, xp, yp, a)
        {
            tokenID = tid;
            tfixed = false;
            this.rgba[0] = this.rgba[1] = this.rgba[2] = this.rgba[3] = -1;
        }

        /**
         * <summary>
         * This constructor takes the provided Session ID, Symbol ID, X and Y coordinate
         * and angle, and assigs these values to the newly created TuioObject.</summary>
         *
         * <param name="si">the Session ID to assign</param>
         * <param name="sym">the Symbol ID to assign</param>
         * <param name="tid">the name ID to assign</param>
         * <param name="xp">the X coordinate to assign</param>
         * <param name="yp">the Y coordinate to assign</param>
         * <param name="a">the angle to assign</param>
         */
        public TuioWallToken(long si, int sym, string tid, float xp, float yp, float a)
            : base(si, sym, xp, yp, a)
        {
            tokenID = tid;
            tfixed = false;
            this.rgba[0] = this.rgba[1] = this.rgba[2] = this.rgba[3] = -1;
        }

        /**
         * <summary>
         * This constructor takes the atttibutes of the provided TuioObject
         * and assigs these values to the newly created TuioObject.</summary>
         *
         * <param name="tobj">the TuioObject to assign</param>
         * <param name="tid">the name ID to assign</param>
         */
        public TuioWallToken(TuioObject tobj, string tid)
            : base(tobj)
        {
            tokenID = tid;
            tfixed = false;
            this.rgba[0] = this.rgba[1] = this.rgba[2] = this.rgba[3] = -1;
        }
        #endregion

        #region Properties & Getter/Setter Methods

        /**
         * <summary>
         * Returns the Token ID  name of this TuioWallToken.</summary>
         * <returns>the Token ID of this TuioWallToken</returns>
         */
        public string TokenID
        {
            get { return tokenID; }
        }

        public bool Fixed
        {
            get { return tfixed; }
            set { this.tfixed = value; }
        }

        public int[] RGBA
        {
            get { return rgba; }
            set { this.rgba = value; }
        }

        public string Shape
        {
            get { return shape; }
            set { this.shape = value; }
        }
        public float Ray
        {
            get { return ray; }
            set { this.ray = value; }
        }
        public float Inner
        {
            get { return inner; }
            set { this.inner = value; }
        }
        #endregion
    }

}
