using System;
using System.Threading;
using System.Net;
using System.Net.Sockets;
using System.Collections.Generic;
using System.IO;
using TouchTokens;
using System.Diagnostics;

using TUIO;
using UnityOSC;

namespace WTT
{
	class WallTokensTuio : TuioListener
	{
 		// Screensize in MM
 		private static float screenWidthMM = 5900;
 		private static float screenHeightMM = 1950;

 		private static int START_TOKEN_SYMB = 7000;

 		private static int MODE_NONE = 0;
 		private static int MODE_OBJECT = 1;
 		private static int MODE_CURSOR = 2;
 		private static int MODE_TOKEN = 3;
 		private static int MODE_MAX = 3;

 		private int mode  = MODE_NONE;

 		// --------------------------------------------------------
 		// some Inner class that might become external or not ...
 		class InnerTuioObject {
 			public TuioObject obj;
 			public long localSessionID;
 			public InnerTuioObject(TuioObject o, long si){
 				obj = o;
 				localSessionID = si;
 			}
 		}
 		class InnerTuioBlob {
 			public TuioBlob blob;
 			public long localSessionID;
 			public InnerTuioBlob(TuioBlob b, long si){
 				blob = b;
 				localSessionID = si;
 			}
 		}
 		class InnerTuioCursor{
 			public TuioCursor c;
 			public long localSessionID;
 			public TouchPoint tp; // in mm !
 			public TuioPoint downtp; // in Tuio
 			public long downTime;
 			public bool inToken;
 			public bool suctionCup;
 			public bool missing;
 			public bool dirty;

 			public InnerTuioCursor(TuioCursor _c, long si = -1){
 				c = _c;
 				localSessionID = si;
 				downTime = _c.TuioTime.TotalMilliseconds;
 				inToken = false;
 				suctionCup = false;
 				tp = new TouchPoint(c.X*screenWidthMM, c.Y*screenHeightMM);
 				downtp = new TuioPoint(_c.TuioTime, _c.X, _c.Y);
 				missing = false;
 				dirty = true;
 			}
 			public InnerTuioCursor(TuioCursor _c, long si, long dt){
 				c = _c;
 				localSessionID = si;
 				downTime = dt;
 				inToken = false;
 				suctionCup = false;
 				tp = new TouchPoint(c.X*screenWidthMM, c.Y*screenHeightMM);
 				downtp = new TuioPoint(_c.TuioTime, _c.X, _c.Y);
 				dirty = true;
 			}
 			public InnerTuioCursor(TuioCursor _c, long dt, bool it) : this(_c, dt)
 			{
 				inToken = it;
 			}
 			public void Update(TuioCursor _c){
 				c = _c;
 				tp.x = _c.X*screenWidthMM;
 				tp.y = _c.Y*screenHeightMM;
 			}
 			public void Update(float x, float y){
 				c.update(x, y); // FIXME time etc
 				tp.x = x*screenWidthMM;
 				tp.y = y*screenHeightMM;
 			}
 			public double Distance(InnerTuioCursor ic){
 				return tp.distance(ic.tp);
 			}
 		} // InnerTuioCursor

 		// --------------------------------------------------------
 		class InnerTuioToken{
 			public TuioObject obj;
 			public TuioCursor cur;
 			public TuioBlob blob;
 			public long localSessionID;
 			public TokenTemplate template = null;
 			public TuioWallToken twk;
 			public List<InnerTuioCursor> ics = new List<InnerTuioCursor>();
 			public TouchPoint centroidPoint;
 			public TuioCursor suctionCupCursor = null;
 			public TouchPoint suctionCupPoint = null;
 			public float angle;
 			public bool oriented;
 			public bool dirty;
 			private TouchPoint refPointAngle = new TouchPoint(50, 0);
 			private List<InnerTuioCursor> downPoints = new List<InnerTuioCursor>();
 			private int numDownPoints;

 			private void ComputeCentroid(){
 				 List<TouchPoint> tmp = new List<TouchPoint>();
 				 foreach(InnerTuioCursor ic in ics){
 				 	tmp.Add(ic.tp);
 				 }
 				 centroidPoint = TouchTokenRecognizer.centroid(tmp);
 			}

 			public  InnerTuioToken(long si, int sym, TokenTemplate _template,  List<InnerTuioCursor> _ics){
 				template = _template;
 				localSessionID = si;
 				ics = new  List<InnerTuioCursor>();
 				SetupPoints(_ics);
 				ComputeCentroid();
 				ComputeAngle();
 				numDownPoints = 3;
 				obj = new TuioObject(si, sym, (float)(centroidPoint.x/screenWidthMM), (float)(centroidPoint.y/screenHeightMM), angle);
 				cur  = new TuioCursor(si, sym, (float)(centroidPoint.x/screenWidthMM), (float)(centroidPoint.y/screenHeightMM));
 				twk = new TuioWallToken(obj, template.TokenID);
 				twk.RGBA = template.RGBA;
 				twk.Shape = template.Shape;
 				twk.Ray = (float)template.Ray;
 				twk.Inner = (float)template.Inner;
 				blob = new TuioBlob(si, sym, (float)(centroidPoint.x/screenWidthMM), (float)(centroidPoint.y/screenHeightMM), angle, twk.Ray, twk.Ray, twk.Ray*twk.Ray);
 			}

 			private void ComputeAngle(){
 				angle = (float)TouchTokenRecognizer.angleBetweenVectors(
 					new TouchPoint(ics[0].tp.x - centroidPoint.x, ics[0].tp.y - centroidPoint.y),
 					refPointAngle
 					);
 			}

 			// order: 1st point the nose
 			private  void SetupPoints(List<InnerTuioCursor> _ics){
 				int n=0;
 				double maxdist = 0;
 				double prevmaxdist = 0;
 				for (int i=0; i < _ics.Count; i++){
 					double dist = 0;
 					for (int j=0; j < _ics.Count; j++){
 						if (i==j){ continue; }
 						dist = dist + _ics[i].tp.distance(_ics[j].tp);
 					}
 					if(dist > maxdist){
 						n = i;
 						prevmaxdist = dist;
 						maxdist = dist;
 					}
 					else if (dist > prevmaxdist){
 						prevmaxdist = dist;
 					}
 				}
 				InnerTuioCursor refPoint = _ics[n];
 				oriented = false;
 				if (maxdist > prevmaxdist + 7){ // FIXME 7
 					oriented = true;
 				}
 				ics.Clear();
 				downPoints.Clear();
 				ics.Add(refPoint);
 				downPoints.Add(new InnerTuioCursor(new TuioCursor(refPoint.c.TuioTime, -1, -1, refPoint.c.X, refPoint.c.Y)));
 				for (int i=0; i < _ics.Count; i++){
 					if (i != n) { 
 						ics.Add(_ics[i]); 
 						downPoints.Add(new InnerTuioCursor(
 							new TuioCursor(_ics[i].c.TuioTime, -1, -1, _ics[i].c.X, _ics[i].c.Y))); 
 					}
 				}

 			}

 			private void updateMissing(float dx, float dy){
 				if (numDownPoints >= 3){ return; }
 				//Console.WriteLine("UPDATE MISSING: "+ twk.TokenID +" "+numDownPoints);
 				if (numDownPoints == 1){
 					for(int i = 0; i < ics.Count; i++){
 						if (ics[i].missing){
 							ics[i].Update(ics[i].c.X + dx, ics[i].c.Y + dy);
 						}
 					}
 					return;
 				}
 				// 2 points 
 				int n1 = 0, n2 = 0, m = 0, ct = 0;
 				for(int i = 0; i < ics.Count; i++){
 					if (!ics[i].missing){
 						if (ct == 0){
 							ct++; n1 = i;
 						}
 						else {
 							n2 = i;
 						}
 					}
 					else {
 						m = i;
 					}
 				}
 				// token rotation vs original token
 				float a1 = (float)TouchTokenRecognizer.angleBetweenVectors(
 					new TouchPoint(ics[n2].tp.x - ics[n1].tp.x, ics[n2].tp.y - ics[n1].tp.y),
 					new TouchPoint(downPoints[n2].tp.x - downPoints[n1].tp.x, downPoints[n2].tp.y - downPoints[n1].tp.y)
 				);
 				//Console.WriteLine("ANGLE TOK: "+ (a1*180/Math.PI));
 				// vector of the missing point for the original token
 				double x = downPoints[m].tp.x - downPoints[n1].tp.x;
 				double y = downPoints[m].tp.y - downPoints[n1].tp.y;
 				// rotate this vector
 				double rx = x*Math.Cos(a1) - y*Math.Sin(a1);
 				double ry = x*Math.Sin(a1) + y*Math.Cos(a1);
 				//Console.WriteLine("MISSING VECT: "+ x +" "+ y +" new: "+ rx +" "+ ry);
 				// translate back
 				rx = rx + downPoints[n1].tp.x;
 				ry = ry + downPoints[n1].tp.y;
 				//Console.WriteLine("MISSING POINT: "+ downPoints[m].tp.x +" "+ downPoints[m].tp.y +" new: "+ rx +" "+ ry);
 				// move back to token
 				rx = rx + (ics[n1].tp.x - downPoints[n1].tp.x);
 				ry = ry + (ics[n1].tp.y - downPoints[n1].tp.y);
 				//Console.WriteLine("MISSING POINT F: "+rx +" "+ ry +" old: " + ics[m].tp.x  +" "+ ics[m].tp.y);
				//Console.WriteLine("MISSING POINT F: "+((float)rx/screenWidthMM)+" "+(float)(ry/screenHeightMM));
 				// translate to the real tok and vect n1 -> m
				x = downPoints[m].tp.x + (ics[n1].tp.x - downPoints[n1].tp.x);
 				y = downPoints[m].tp.y + (ics[n1].tp.y - downPoints[n1].tp.y);
 				// vector
 				x = x - ics[n1].tp.x;
 				y = y - ics[n1].tp.y;
 				// rotate
 				//a1 = -a1;
 				rx = x*Math.Cos(a1) - y*Math.Sin(a1);
 				ry = x*Math.Sin(a1) + y*Math.Cos(a1);
 				// vect to point
 				rx = rx + ics[n1].tp.x;
 				ry = ry + ics[n1].tp.y;

 				ics[m].Update((float)(rx/screenWidthMM), (float)(ry/screenHeightMM));
 			}

 			public bool Update(TuioCursor _c){
 				// todo angle
 				bool updated = false;
 				float dx = 0, dy = 0;
 				foreach(InnerTuioCursor ic in ics){
 					if (ic.c.SessionID == _c.SessionID){
 						dx = _c.X - ic.c.X;
 						dy = _c.Y - ic.c.Y;
 						ic.Update(_c.X, _c.Y); // time etc.
 						updated = true;
 						break;
 					}
 				}
 				if (updated) {
 					updateMissing(dx, dy);
 					ComputeCentroid();
 					ComputeAngle();
 					obj.update(_c.TuioTime, (float)(centroidPoint.x/screenWidthMM), (float)(centroidPoint.y/screenHeightMM), angle);
 					blob.update(_c.TuioTime, (float)(centroidPoint.x/screenWidthMM), (float)(centroidPoint.y/screenHeightMM), angle, twk.Ray, twk.Ray, twk.Ray*twk.Ray);
 					cur.update(_c.TuioTime, (float)(centroidPoint.x/screenWidthMM), (float)(centroidPoint.y/screenHeightMM));
 					twk.update(_c.TuioTime, (float)(centroidPoint.x/screenWidthMM), (float)(centroidPoint.y/screenHeightMM), angle);
 				}
 				return updated;
 			}

 			public int Remove(TuioCursor _c){
 				InnerTuioCursor ic_torm = null;
 				foreach(InnerTuioCursor ic in ics){
 					if (ic.c.SessionID == _c.SessionID){
 						ic_torm = ic;
 						break;
 					}
 				}
 				if (ic_torm != null) { 
 					//ics.Remove(ic_torm);
 					numDownPoints--;
 					ic_torm.missing = true;
 					// FIXME: angle blob
 					if (numDownPoints > 0){
 						ComputeCentroid();
 						obj.update(_c.TuioTime, (float)(centroidPoint.x/screenWidthMM), (float)(centroidPoint.y/screenHeightMM), angle);
 						cur.update(_c.TuioTime, (float)(centroidPoint.x/screenWidthMM), (float)(centroidPoint.y/screenHeightMM));
 						twk.update(_c.TuioTime, (float)(centroidPoint.x/screenWidthMM), (float)(centroidPoint.y/screenHeightMM), angle);
 					}
 				}
 				return numDownPoints;
 			}

 			public void AddMissingCursor(InnerTuioCursor ic){
 				// FIXME BROKEN
 				// FIXME 2 missing
 				// FIXME blob
 				for (int i = 0; i < ics.Count; i++){
 					if (ics[i].missing){
 						ics[i] = new InnerTuioCursor(ic.c, ic.localSessionID);
 					}
 				}
 				ComputeCentroid();
				ComputeAngle();
 				obj.update(ic.c.TuioTime, (float)(centroidPoint.x/screenWidthMM), (float)(centroidPoint.y/screenHeightMM), angle);
 				cur.update(ic.c.TuioTime, (float)(centroidPoint.x/screenWidthMM), (float)(centroidPoint.y/screenHeightMM));
 				twk.update(ic.c.TuioTime, (float)(centroidPoint.x/screenWidthMM), (float)(centroidPoint.y/screenHeightMM), angle);
 			}

 			public void setSuctionCup(InnerTuioCursor ic){
 				suctionCupCursor = ic.c;
 				suctionCupPoint = ic.tp;
 				twk.Fixed = true;
 			}

 			public void removeSuctionCup(){
 				suctionCupCursor = null;
 				suctionCupPoint = null;
 				twk.Fixed = false;
 			}
 		} // InnerTuioToken

 		// ---------------------------------------------------------
 		// stuff

 		// recieving from TUIO
 		private TuioClient client;

		private bool debugStart = true;
		private bool debugAddRemove = true;
		private bool debugCursorUpdate = false;
		private bool debugTokenUpdate = false;
		private bool debugObjectUpdate = false;
		private bool debugBlobUpdate = false;

		// sending to a TUIO Client ()
		private  OSCClient sender = null;
		// and/or to TuioListener
		private List<TuioListener> listenerList = new List<TuioListener>();
 		// and/or to TuioListener
		private List<TuioWallTokenListener> wallTokenListenerList = new List<TuioWallTokenListener>();

		// token
		private TouchTokenRecognizer recognizer;

		// Tokens currently on the wall
		private List<InnerTuioToken> tokens = new List<InnerTuioToken>();

		// cursors
		private List<InnerTuioCursor> touchCursors = new List<InnerTuioCursor>(); // we send events with these one
		private List<InnerTuioCursor> waitCursors = new List<InnerTuioCursor>();  // we wait for these ones
		//
		private List<InnerTuioObject> iTuioObjects = new List<InnerTuioObject>(); // external blob
		private List<InnerTuioBlob> iTuioBlobs = new List<InnerTuioBlob>(); // external object

		// should have our own session id
		private long lastLocalSessionID = 0;

		//
		private int _inPort = 3333;
		private string _templateFileName;
		private string _outIP = "127.0.0.1";
		private int _outPort = 0;

		// logger
		private StreamWriter logWriter;
		private Stopwatch logwatch = new Stopwatch();


		// recognizer param
		private long waitCursorsDelay = 150; // ms or 200 ?
		private float minSquaredDistTokenPairsTolerance = 30000000;
		private float toleranceRatio = 2f;

		// standard + WallTokenTuioListener 
		public WallTokensTuio(string templateFileName, float swmm, float shmm){
			_templateFileName = templateFileName;
			screenWidthMM = swmm;
			screenHeightMM = shmm;
		}
		public WallTokensTuio(int inPort, string templateFileName, float swmm, float shmm) : this(templateFileName, swmm, shmm)
		{
			_inPort = inPort;
		}

		// fwd + walltoken tuio
		public WallTokensTuio(string templateFileName, float swmm, float shmm, int outPort) : this(templateFileName, swmm, shmm)
		{
			_outPort = outPort;
		}
		public WallTokensTuio(string templateFileName, float swmm, float shmm, int outPort, string outIP) : this(templateFileName, swmm, shmm, outPort)
		{
			_outIP = outIP;
		}
		public WallTokensTuio(int inPort, string templateFileName, float swmm, float shmm, int outPort) : this(templateFileName, swmm, shmm, outPort)
		{
			_inPort = inPort;
		}
		public WallTokensTuio(int inPort, string templateFileName, float swmm, float shmm, int outPort, string outIP) : this(inPort, templateFileName, swmm, shmm, outPort)
		{
			_outIP = outIP;
		}

		// fwd + pure Tuio
		public WallTokensTuio(string templateFileName, float swmm, float shmm, int outPort, int m) : this(templateFileName, swmm, shmm, outPort)
		{
			mode = m;
			if (mode < 0 || mode > MODE_MAX) { mode = MODE_OBJECT; }
		}
		public WallTokensTuio(string templateFileName, float swmm, float shmm, int outPort, string outIP, int m)  : this(templateFileName, swmm, shmm, outPort, m)
		{
			_outIP = outIP;
		}
		public WallTokensTuio(int inPort, string templateFileName, float swmm, float shmm, int outPort, int m) : this(inPort, templateFileName, swmm, shmm, outPort)
		{
			mode = m;
			if (mode < 0 || mode > MODE_MAX) { mode = MODE_OBJECT; }
		}
		public WallTokensTuio(int inPort, string templateFileName, float swmm, float shmm, int outPort, string outIP, int m) : this(inPort, templateFileName, swmm, shmm, outPort, m)
		{
			_outIP = outIP;
		}

		private bool started = false;

		public void Start()
		{
			// in going
			client = new TuioClient(_inPort);
			client.addTuioListener(this);
			if (debugStart){
				Console.WriteLine("Create TUIO clien " + _inPort + " " + client);
			}
			// out going
			if (_outPort > 0){
				sender = new OSCClient(IPAddress.Parse(_outIP), _outPort);
				if (debugStart){
					Console.WriteLine("Get an out port, sender: " + _outPort + " " + _outIP + " " + sender);
				}
			}

			recognizer = new TouchTokenRecognizer(_templateFileName);
			if (debugStart){
				Console.WriteLine("Num of token template: "+recognizer.tokenTemplates.Count+
					" ("+ _templateFileName +")");
			}
			double minDist = double.MaxValue;
			foreach(TokenTemplate tplt in recognizer.tokenTemplates){
				if (debugStart) {
					Console.WriteLine(tplt.TokenID);
				}
				if (logWriter != null){
					logWriter.WriteLine(tplt.ToString());
				}
				List<TouchPoint> tPoints = TouchTokenRecognizer.orderAndAlign(tplt.Points, TouchTokenRecognizer.centroid(tplt.Points));
				foreach(TokenTemplate tplt2 in recognizer.tokenTemplates){
					List<TouchPoint> input = TouchTokenRecognizer.orderAndAlign(tplt2.Points, TouchTokenRecognizer.centroid(tplt2.Points));
					double d = TouchTokenRecognizer.minDistance(tPoints, input, TouchTokenRecognizer.centroid(input));
					if (tplt2.TokenID != tplt.TokenID && d < minDist){
						minDist = d;
					}
					if (logWriter != null){
						logWriter.WriteLine("\t distance with " +tplt2.TokenID+"\t"+ d);
					}
				}
			}
			
			minSquaredDistTokenPairsTolerance = (float)minDist/toleranceRatio;
			if (debugStart){
				Console.WriteLine("Min dist between tokens: "+ minDist+ " tolerance: " + minSquaredDistTokenPairsTolerance);
			}
			if (logWriter != null){
				logWriter.WriteLine("Min dist between tokens: "+ minDist + " tolerance: " + minSquaredDistTokenPairsTolerance);
				logWriter.Flush();
			}

			if (debugStart){
				Console.WriteLine("Connect to client " + _inPort + " " + client);
			}
			try {
				client.connect();
			}
			catch(Exception e)
			{
				Console.WriteLine("\n !!!! WallTokensTuio: Fail to connect "+ _inPort + " "+ e);
			}

			List<TuioCursor> curList = client.getTuioCursors();
			long maxSessionID = 0;
			foreach(TuioCursor c in curList){
				if (c.SessionID >= maxSessionID){
					maxSessionID = c.SessionID;
				}
				touchCursors.Add(new InnerTuioCursor(c, c.SessionID));
			}
			List<TuioObject> objList = client.getTuioObjects();
			foreach(TuioObject o in objList){
				if (o.SessionID >= maxSessionID){
					maxSessionID = o.SessionID;
				}
				iTuioObjects.Add(new InnerTuioObject(o, o.SessionID));
			}
			List<TuioBlob> blobList = client.getTuioBlobs();
			foreach(TuioBlob b in blobList){
				if (b.SessionID >= maxSessionID){
					maxSessionID = b.SessionID;
				}
				iTuioBlobs.Add(new InnerTuioBlob(b, b.SessionID));
			}
			lastLocalSessionID = maxSessionID;
			if (debugStart){
				Console.WriteLine("our last local session id: "+lastLocalSessionID);
				Console.WriteLine("STARTED!");

			}

			logwatch.Start();
			started = true;
		}

		public void SetVerbosity(int v){
			debugStart = false;
			debugAddRemove = false;
			debugCursorUpdate = false;
			debugTokenUpdate = false;
			debugObjectUpdate = false;
			debugBlobUpdate = false;
			
			if (v >= 1){
				debugStart = true;
			}
			if (v >= 2){
				debugAddRemove = true;
			}
			if (v >= 3){
				debugCursorUpdate = debugTokenUpdate = debugObjectUpdate = debugBlobUpdate = true;
			}
		}

		public void StartLog(string basePathName){
			if (logWriter != null) StopLog();
			string mydate = System.DateTime.Now.ToString("yyyy-MM-dd-HH-mm-ss");
        	string path = basePathName + "-" +  mydate + ".txt";
        	logWriter = new StreamWriter(path, false);
        	logWriter.WriteLine("START "+ logwatch.ElapsedMilliseconds);
		}
		public void StopLog(){
			if (logWriter == null) return;
			logWriter.WriteLine("STOP "+ logwatch.ElapsedMilliseconds);
			logWriter.Flush();

			logWriter = null;
		}

		// override what it is computed here: minSquaredDistTokenPairsTolerance
		public float ToleranceRatio
        {
            get { return toleranceRatio; }
            set { this.toleranceRatio = value; }
        }

		public List<TokenTemplate> GetAllTokenTemplates(){
			if (!started){
				Start();
			}
			if (recognizer == null) { return null; }
			return recognizer.tokenTemplates;
		}

		public List<string> GetAllTokenIDs(){
			if (!started){
				Start();
			}
			if (recognizer == null) { return null; }
			List<string> ret = new List<string>();
			foreach(TokenTemplate tplt in recognizer.tokenTemplates){
				ret.Add(tplt.TokenID);
			}
			return ret;
		}

		 #region TUIO Listener Management
        /**
		 * <summary>
         * Adds the provided TuioListener to the list of registered TUIO event listeners</summary>
         * <param name="listener">the TuioListener to add</param>
		 */
        public void AddTuioListener(TuioListener listener)
        {
            listenerList.Add(listener);
        }

        /**
         * <summary>
         * Removes the provided TuioListener from the list of registered TUIO event listeners</summary>
         * <param name="listener">the TuioListener to remove</param>
         */
        public void RemoveTuioListener(TuioListener listener)
        {
            listenerList.Remove(listener);
        }

        /**
         * <summary>
         * Removes all TuioListener from the list of registered TUIO event listeners</summary>
         */
        public void RemoveAllTuioListeners()
        {
            listenerList.Clear();
        }
        #endregion

         #region Wall Token TUIO Listener Management
        /**
		 * <summary>
         * Adds the provided TuioWallTokenListener to the list of registered TUIO event listeners</summary>
         * <param name="listener">the TuioWallTokenListener to add</param>
		 */
        public void AddTuioWallTokenListener(TuioWallTokenListener listener)
        {
            wallTokenListenerList.Add(listener);
        }

        /**
         * <summary>
         * Removes the provided TuioWallTokenListener from the list of registered TUIO event listeners</summary>
         * <param name="listener">the TuioWallTokenListener to remove</param>
         */
        public void RemoveTuioWallTokenListener(TuioWallTokenListener listener)
        {
            wallTokenListenerList.Remove(listener);
        }

        /**
         * <summary>
         * Removes all TuioWallTokenListener from the list of registered TUIO event listeners</summary>
         */
        public void RemoveAllTuioWallTokenListeners()
        {
            wallTokenListenerList.Clear();
        }
        #endregion

        #region Send Msg 
		private void SendMessage<T>(string address, List<T> values)
		{	
			OSCMessage message = new OSCMessage(address);
		
			foreach(T msgvalue in values)
			{
				message.Append(msgvalue);
			}
			message.Pack();
			try 
			{
				sender.Send(message);
			}
			catch
			{
				//Console.WriteLine(String.Format("Can't send OSC packet to client {0} : {1}", 0, 0));
			}
		}

		private int frameID = 0;
		private void SendFrameID(string address){
				frameID++;
				List<object> args = new List<object>();
				args.Add("fseq");
				args.Add(frameID);
				try 
				{
					SendMessage(address, args);
				}
				catch
				{
					//Console.WriteLine(String.Format("Can't send OSC packet to client {0} : {1}", 0, 0));
				}
		}
		#endregion

		#region Utils for recognition
		private float distance(float x1, float y1, float x2,float y2){
			return (float)Math.Pow(Math.Pow(x1-x2,2) + Math.Pow(y1-y2,2), 0.5);
		}

		private float DistanceBetweenCursorMM(TuioCursor c1,  TuioCursor c2){
			float x1 = screenWidthMM*c1.X;
			float y1 = screenHeightMM*c1.Y;
			float x2 = screenWidthMM*c2.X;
			float y2 = screenHeightMM*c2.Y;
			return distance(x1,y1,x2,y2);
		}

		private float DistanceBetweenCursorsMM(TuioCursor c,  List<TuioCursor> cl){
			float dist = -1;
			foreach (TuioCursor t in cl){
				float d = DistanceBetweenCursorMM(c,  t);
				if (dist == -1 || dist > d){
					dist = d;
				}
			}
			return dist;
		}

		double sign (TouchPoint p1, TouchPoint p2, TouchPoint p3)
		{
    		return (p1.x - p3.x) * (p2.y - p3.y) - (p2.x - p3.x) * (p1.y - p3.y);
		}

		bool PointInTriangle (TouchPoint pt, TouchPoint v1, TouchPoint v2, TouchPoint v3)
		{
    		double d1, d2, d3;
    		bool has_neg, has_pos;

    		d1 = sign(pt, v1, v2);
    		d2 = sign(pt, v2, v3);
    		d3 = sign(pt, v3, v1);

    		has_neg = (d1 < 0) || (d2 < 0) || (d3 < 0);
    		has_pos = (d1 > 0) || (d2 > 0) || (d3 > 0);

    		return !(has_neg && has_pos);
		}

		bool PointUnderToken(TouchPoint pt, InnerTuioToken it)
		{
				// FIXME 2
    		return (Math.Pow(it.centroidPoint.x - pt.x, 2) + Math.Pow(it.centroidPoint.y - pt.y, 2) < Math.Pow(2*it.twk.Ray, 2));
		}
		#endregion

		// ---------------------------------------------------
		#region TUIO Cursors
		private void ReallyAddTuioCursor(InnerTuioCursor ic, bool isToken = false){
			// fwd as touch, indeed
			if (isToken && mode != MODE_CURSOR) { return; }

			if (!isToken) {
				touchCursors.Add(ic);
			}

			// should add
			TuioCursor cc = new TuioCursor(ic.downtp.StartTime, ic.c.SessionID, ic.c.CursorID, ic.downtp.X, ic.downtp.Y);
			cc.update(ic.c);
			if (debugAddRemove){
				Console.WriteLine("Add Real Cursor: "+ isToken +" "+ ic.localSessionID +" x: "+cc.X +" y: "+ cc.Y);
			}
			for (int i = 0; i < listenerList.Count; i++)
			{
				TuioListener listener = (TuioListener)listenerList[i];
				if (listener != null) listener.addTuioCursor(cc);
			}
			if (!isToken) {
				for (int i = 0; i < wallTokenListenerList.Count; i++)
				{
					TuioWallTokenListener listener = (TuioWallTokenListener)wallTokenListenerList[i];
					if (listener != null) listener.addTuioCursor(cc);
				}
			}

			//
			for (int i = 0; i < listenerList.Count; i++)
			{
				TuioListener listener = (TuioListener)listenerList[i];
				if (listener != null) listener.updateTuioCursor(ic.c);
			}
			if (!isToken) {
				for (int i = 0; i < wallTokenListenerList.Count; i++)
				{
					TuioWallTokenListener listener = (TuioWallTokenListener)wallTokenListenerList[i];
					if (listener != null) listener.updateTuioCursor(ic.c);
				}
			}

			if (sender == null){
				return;
			}

			List<object> args = new List<object>();

			// args.Clear();
			// args.Add("alive");
			// foreach(InnerTuioCursor tic in touchCursors){
			// 	args.Add((int)tic.localSessionID);
			// }
			// SendMessage("/tuio/2Dcur", args);
			// SendFrameID("/tuio/2Dcur");

			args.Clear();
			args.Add("set");
			args.Add((int)ic.localSessionID);
			args.Add((float)cc.X);
			args.Add((float)cc.Y);
			args.Add((float)ic.c.XSpeed); // xspeed
			args.Add((float)ic.c.YSpeed); // yspeed
			args.Add((float)ic.c.MotionAccel); // maccel
			
			SendMessage("/tuio/2Dcur", args);
			SendFrameID("/tuio/2Dcur");

			args.Clear();
			args.Add("alive");
			foreach(InnerTuioCursor tic in touchCursors){
				args.Add((int)tic.localSessionID);
			}
			if (mode == MODE_CURSOR){
				foreach (InnerTuioToken it in tokens){
					args.Add((int)it.localSessionID);
				}
			}
			SendMessage("/tuio/2Dcur", args);
			SendFrameID("/tuio/2Dcur");

			args.Clear();
			args.Add("set");
			args.Add((int)ic.localSessionID);
			args.Add((float)ic.c.X);
			args.Add((float)ic.c.Y);
			args.Add((float)ic.c.XSpeed); // xspeed
			args.Add((float)ic.c.YSpeed); // yspeed
			args.Add((float)ic.c.MotionAccel); // maccel
			
			SendMessage("/tuio/2Dcur", args);
			SendFrameID("/tuio/2Dcur");

		}

		public void addTuioCursor(TuioCursor c)
		{
			lastLocalSessionID++;
			InnerTuioCursor addedIC = new InnerTuioCursor(c, lastLocalSessionID);

			if (logWriter != null){
				logWriter.WriteLine("ADD_CURSOR " + logwatch.ElapsedMilliseconds + " " + c.TuioTime.TotalMilliseconds + " " +
					c.SessionID + " " + c.X + " " + c.Y);
			}

			if (c.X < 0 || c.Y < 0 || c.X > 1 || c.Y > 1) { 
				return; // ignore
			}
			if (c.X == 0 && c.Y == 0){
				return;
			}
			//Re-attach token foot (NOT MANAGED WITH CLOSE TOKENS)
			//If there is at least one token on the wall
			// FIXME: do this in refresh ? FIXME !!!
			InnerTuioToken iToken = null;
			InnerTuioToken gToken = null; // ghost
			foreach (InnerTuioToken it in tokens){
				if (it.ics.Count == 2){
					List<TouchPoint> tmp = new List<TouchPoint>();
					foreach(InnerTuioCursor ic in it.ics)
					{
						tmp.Add(ic.tp);
					}
					tmp.Add(addedIC.tp);
					TokenTemplate t = recognizer.recognize(tmp);
					if (t != null && t.TokenID == it.template.TokenID && t.Distance <= minSquaredDistTokenPairsTolerance){
						// ok !
						iToken = it;
						break;
					}
				}
				else if (it.ics.Count == 1){
					// FIXME more difficult :)
				}
			}
			if (iToken != null){
				iToken.AddMissingCursor(addedIC);
				// FIXME Generate an event ?
				return;
			}

			// now check if it is a suctionCupPoint
			foreach (InnerTuioToken it in tokens){
				if (it.ics.Count < 3){
					continue; // I guess
				}
				if (it.suctionCupPoint != null){
					if (PointUnderToken(addedIC.tp, it)) {
						gToken = it;
						break;
					}
					continue;
				}
				if (PointInTriangle(addedIC.tp, it.ics[0].tp,  it.ics[1].tp,  it.ics[2].tp)) {
					iToken = it;
					break;
				}
			}
			if (iToken != null){
				iToken.setSuctionCup(addedIC);
				SetTuioToken(iToken, false);
				return;
			}
			if (gToken != null){
				return; // ignore !!
			}
			// free the waitCursors
			List<InnerTuioCursor> tmpics = new List <InnerTuioCursor>();
			foreach(InnerTuioCursor ic in waitCursors){
				if (addedIC.downTime - ic.downTime > waitCursorsDelay){
					tmpics.Add(ic);
				}
			}
			foreach(InnerTuioCursor ic in tmpics){
				ReallyAddTuioCursor(ic);
				waitCursors.Remove(ic);
			}

			waitCursors.Add(addedIC);
		}


		private void SetTuioCursor(InnerTuioCursor ic, bool isToken = false){
			if (isToken && mode != MODE_CURSOR) { return; }
			for (int i = 0; i < listenerList.Count; i++)
			{
				TuioListener listener = (TuioListener)listenerList[i];
				if (listener != null) listener.updateTuioCursor(ic.c);
			}
			if (!isToken) {
				for (int i = 0; i < wallTokenListenerList.Count; i++)
				{
					TuioWallTokenListener listener = (TuioWallTokenListener)wallTokenListenerList[i];
					if (listener != null) listener.updateTuioCursor(ic.c);
				}
			}
			if (sender == null){
				return;
			}
			List<object> args = new List<object>();
			args.Add("set");
			args.Add((int)ic.localSessionID);
			args.Add((float)ic.c.X);
			args.Add((float)ic.c.Y);
			args.Add((float)ic.c.XSpeed); // xspeed
			args.Add((float)ic.c.YSpeed); // yspeed
			args.Add((float)ic.c.MotionAccel); // maccel
			
			SendMessage("/tuio/2Dcur", args);
			SendFrameID("/tuio/2Dcur");
		}

		public void updateTuioCursor(TuioCursor c)
		{
			InnerTuioCursor updateIC = new InnerTuioCursor(c, -1);

			if (logWriter != null){
				logWriter.WriteLine("UPDATE_CURSOR " + logwatch.ElapsedMilliseconds + " " + c.TuioTime.TotalMilliseconds + " " +
					c.SessionID + " " + c.X + " " + c.Y);
			}

			// waiting cursors
			bool isInWaitList = false;

			// waitCursor
			List<InnerTuioCursor> tmpics = new List <InnerTuioCursor>();
			foreach(InnerTuioCursor ic in waitCursors){
				if (ic.c.SessionID == updateIC.c.SessionID){
					isInWaitList = true;
					//if ( (updateIC.Distance(ic) >= 10) || updateIC.downTime - ic.downTime > waitCursorsDelay) {
					if ( updateIC.downTime - ic.downTime > waitCursorsDelay) {
						// 10 mm or waitCursorsDelay ms
						// 10 mm FIXME might be a good
						tmpics.Add(ic);
						isInWaitList = false;
						continue;
					}
					ic.Update(c);

				}
				if (updateIC.downTime - ic.downTime > waitCursorsDelay){
					tmpics.Add(ic);
				}
				
			}
			foreach(InnerTuioCursor ic in tmpics){
				ReallyAddTuioCursor(ic);
				waitCursors.Remove(ic);
			}
			if (isInWaitList){
				return;
			}

			// touch cursors
			foreach(InnerTuioCursor ic in touchCursors){
				if (ic.c.SessionID != updateIC.c.SessionID){ continue; }
				ic.Update(c);
				if (debugCursorUpdate){
					Console.WriteLine("Update touch Cursor: "+ic.localSessionID +" x: "+ic.c.X +" y: "+ ic.c.Y);
				}
				SetTuioCursor(ic);
				break;
			}

			// tokens
			for(int i = 0; i < tokens.Count; i++){
				InnerTuioToken it = tokens[i];
				if (it.Update(c)){
					it.dirty = true;
					if (debugTokenUpdate){
						Console.WriteLine("Update Token: " + it.twk.TokenID + " sid: "+ it.twk.SessionID + " sym:"+ it.twk.SymbolID +" x: "+ it.twk.X +" y: "+ it.twk.Y +" a: "+ it.twk.Angle*180/Math.PI);
					}
					// update in referesh
				}
			}
		}

		private void SendRemoveTuioCursor(InnerTuioCursor ic, bool isToken = false){
			if (isToken && mode != MODE_CURSOR) { return; }
			TuioCursor c = ic.c;
			TuioCursor nc = new TuioCursor(c.TuioTime, ic.localSessionID, c.CursorID, c.X, c.Y);
			nc.update(c);
			for (int i = 0; i < listenerList.Count; i++)
			{
				TuioListener listener = (TuioListener)listenerList[i];
				if (listener != null) listener.removeTuioCursor(nc);
			}
			if (!isToken){
				for (int i = 0; i < wallTokenListenerList.Count; i++)
				{
					TuioWallTokenListener listener = (TuioWallTokenListener)wallTokenListenerList[i];
					if (listener != null) listener.removeTuioCursor(nc);
				}
			}
			if (sender == null) {
				return;
			}
			List<object> args = new List<object>();
			args.Add("alive");
			foreach(InnerTuioCursor tic in touchCursors){
				args.Add((int)tic.localSessionID);
			}
			if (mode == MODE_CURSOR){
				foreach (InnerTuioToken it in tokens){
					args.Add((int)it.localSessionID);
				}
			}
			args.Add((int)ic.localSessionID);
			SendMessage("/tuio/2Dcur", args);
			SendFrameID("/tuio/2Dcur");
			
			args.Clear();
			args.Add("alive");
			foreach(InnerTuioCursor tic in touchCursors){
				args.Add((int)tic.localSessionID);
			}
			if (mode == MODE_CURSOR){
				foreach (InnerTuioToken it in tokens){
					args.Add((int)it.localSessionID);
				}
			}
			SendMessage("/tuio/2Dcur", args);
			SendFrameID("/tuio/2Dcur");
		}

		public void removeTuioCursor(TuioCursor c)
		{
			InnerTuioCursor removeIC = new InnerTuioCursor(c, -1);
			if (logWriter != null){
				logWriter.WriteLine("REMOVE_CURSOR " + logwatch.ElapsedMilliseconds + " " + c.TuioTime.TotalMilliseconds + " " +
					c.SessionID + " " + c.X + " " + c.Y);
			}
			// waiting cursors
			InnerTuioCursor icToBeRemoved = null;
			// waitCursor
			List<InnerTuioCursor> tmpics = new List <InnerTuioCursor>();
			foreach(InnerTuioCursor ic in waitCursors){
				if (ic.c.SessionID == removeIC.c.SessionID){
					ReallyAddTuioCursor(ic); // should add and then remove !
					icToBeRemoved = ic;
				}
				else if (removeIC.downTime - ic.downTime > waitCursorsDelay){
					tmpics.Add(ic);
				}
				
			}
			waitCursors.Remove(icToBeRemoved);
			foreach(InnerTuioCursor ic in tmpics){
				ReallyAddTuioCursor(ic);
				waitCursors.Remove(ic);
			}
			// if (icToBeRemoved != null){
			//	return;
			//}

			// touch cursors
			foreach(InnerTuioCursor ic in touchCursors){
				if (ic.c.SessionID == removeIC.c.SessionID){ 
					icToBeRemoved = ic;
				}
			}
			if (icToBeRemoved != null){
				if (debugAddRemove){
					Console.WriteLine("Remove touch Cursor: "+ icToBeRemoved.localSessionID +" x: "+icToBeRemoved.c.X +" y: "+ icToBeRemoved.c.Y);
				}
				touchCursors.Remove(icToBeRemoved);
				SendRemoveTuioCursor(icToBeRemoved);
			}

			if (icToBeRemoved != null){
				return;
			}

			// tokens
			InnerTuioToken itToBeRemoved = null;
			foreach(InnerTuioToken it in tokens){
				if (it.Update(c)){
					int rest = it.Remove(c);
					if (rest > 0){
						SetTuioToken(it, false);
						InnerTuioObject tmpo = new InnerTuioObject(it.obj, it.localSessionID);
						SetTuioObject(tmpo, false, true);
						InnerTuioBlob tmpb = new InnerTuioBlob(it.blob, it.localSessionID);
						SetTuioBlob(tmpb, false, true);
						InnerTuioCursor tmpc = new InnerTuioCursor(it.cur, it.localSessionID);
						SetTuioCursor(tmpc, true);
					}
					else {
						itToBeRemoved = it;
					}
					break;
				}
			}
			if (itToBeRemoved != null){
				tokens.Remove(itToBeRemoved);
				if (debugAddRemove){
					Console.WriteLine("Remove Token: " + itToBeRemoved.template.tokenID + " sid: "+ itToBeRemoved.localSessionID + " sym:"+ itToBeRemoved.obj.SymbolID +" x: "+ itToBeRemoved.obj.X +" y: "+ itToBeRemoved.obj.Y);
				}
				SendRemoveTuioToken(itToBeRemoved);
				InnerTuioObject tmpo = new InnerTuioObject(itToBeRemoved.obj, itToBeRemoved.localSessionID);
				SendRemoveTuioObject(tmpo, true);
				InnerTuioCursor tmpc = new InnerTuioCursor(itToBeRemoved.cur, itToBeRemoved.localSessionID);
				SendRemoveTuioCursor(tmpc, true);
				return;
			}
		}
		#endregion

		// -----------------------------------------------------------------
		#region TUIO Token
		private void SetTuioToken(InnerTuioToken it, bool add){
			for (int i = 0; i < wallTokenListenerList.Count; i++)
			{
				TuioWallTokenListener listener = (TuioWallTokenListener)wallTokenListenerList[i];
				if (add){
					if (listener != null) listener.addTuioWallToken(it.twk);
				}
				else {
					if (listener != null) {
						listener.updateTuioWallToken(it.twk);
					}
				}
			}
			if (sender == null || (mode != MODE_TOKEN)) {
				return;
			}

			List<object> args = new List<object>();
			args.Add("set");
			args.Add((int)it.twk.SessionID);
			args.Add((int)it.twk.SymbolID);
			args.Add((string)it.twk.TokenID);
			args.Add((float)it.twk.X);
			args.Add((float)it.twk.Y);
			args.Add((float)it.twk.Angle);
			// FIXME
			args.Add((float)it.twk.XSpeed);
			args.Add((float)it.twk.YSpeed);
			args.Add((float)it.twk.RotationSpeed);
			args.Add((float)it.twk.MotionAccel);
			args.Add((float)it.twk.RotationAccel);
			// TOKEN EXTENSION
			args.Add((int)((it.twk.Fixed)? 1:0));
			args.Add((int)it.twk.RGBA[0]);
			args.Add((int)it.twk.RGBA[1]);
			args.Add((int)it.twk.RGBA[2]);
			args.Add((int)it.twk.RGBA[3]);
			args.Add((string)it.twk.Shape);
			args.Add((float)it.twk.Ray);
			SendMessage("/tuio/2Dtok", args);
			SendFrameID("/tuio/2Dtok");
		}

		private void SendRemoveTuioToken(InnerTuioToken it) {
			for (int i = 0; i < wallTokenListenerList.Count; i++)
			{
				TuioWallTokenListener listener = (TuioWallTokenListener)wallTokenListenerList[i];
				if (listener != null) listener.removeTuioWallToken(it.twk);
			}
			if (sender == null || mode != MODE_TOKEN){
				return;
			}
			List<object> args = new List<object>();
			args.Add("alive");
			for (int i = 0; i < tokens.Count; i++)
			{
				InnerTuioToken ito = (InnerTuioToken)tokens[i];
				if (ito.localSessionID == it.localSessionID){
					continue; // should not happen ....
				}
				args.Add((int)ito.localSessionID);
			}
			args.Add((int)it.localSessionID);
			SendMessage("/tuio/2Dtok", args);
			SendFrameID("/tuio/2Dtok");

			args.Clear();
			args.Add("alive");
			for (int i = 0; i < tokens.Count; i++)
			{
				InnerTuioToken ito = (InnerTuioToken)tokens[i];
				if (ito.localSessionID == it.localSessionID){
					continue; // should not happen ....
				}
				args.Add((int)ito.localSessionID);
			}
			SendMessage("/tuio/2Dtok", args);
			SendFrameID("/tuio/2Dtok");
		}
		#endregion

		// -----------------------------------------------------------------
		#region TUIO Object
		private void SetTuioObject(InnerTuioObject io, bool add, bool isToken){
			if (isToken && mode != MODE_OBJECT) { return; }
			TuioObject o = io.obj;
			TuioObject no = new TuioObject(o.TuioTime, io.localSessionID, o.SymbolID, o.X, o.Y, o.Angle);
			no.update(o);
			for (int i = 0; i < listenerList.Count; i++)
			{
				TuioListener listener = (TuioListener)listenerList[i];
				if (add){
					if (listener != null) listener.addTuioObject(no);
				}
				else {
					if (listener != null) listener.updateTuioObject(no);
				}
			}
			if (!isToken){
				for (int i = 0; i < wallTokenListenerList.Count; i++)
				{
					TuioWallTokenListener listener = (TuioWallTokenListener)wallTokenListenerList[i];
					if (add){
						if (listener != null) listener.addTuioObject(no);
					}
					else {
						if (listener != null) {
							listener.updateTuioObject(no);
						}
					}
				}
			}
			if (sender == null) {
				return;
			}
			List<object> args = new List<object>();
			args.Add("set");
			args.Add((int)io.localSessionID);
			args.Add((int)io.obj.SymbolID);
			args.Add((float)io.obj.X);
			args.Add((float)io.obj.Y);
			args.Add((float)io.obj.Angle);
			args.Add((float)io.obj.XSpeed);
			args.Add((float)io.obj.YSpeed);
			args.Add((float)io.obj.RotationSpeed);
			args.Add((float)io.obj.MotionAccel);
			args.Add((float)io.obj.RotationAccel);

			SendMessage("/tuio/2Dobj", args);
			SendFrameID("/tuio/2Dobj");
		}

		private void SendRemoveTuioObject(InnerTuioObject io, bool isToken) {
			if (isToken && mode != MODE_OBJECT) { return; }
			TuioObject o = io.obj;
			TuioObject no = new TuioObject(o.TuioTime, io.localSessionID, o.SymbolID, o.X, o.Y, o.Angle);
			no.update(o);
			for (int i = 0; i < listenerList.Count; i++)
			{
				TuioListener listener = (TuioListener)listenerList[i];
				if (listener != null) listener.removeTuioObject(no);
			}
			if (!isToken){
				for (int i = 0; i < wallTokenListenerList.Count; i++)
				{
					TuioWallTokenListener listener = (TuioWallTokenListener)wallTokenListenerList[i];
					if (listener != null) listener.removeTuioObject(no);
				}
			}
			if (sender == null) {
				return;
			}
			List<object> args = new List<object>();
			args.Add("alive");
			for (int i = 0; i < iTuioObjects.Count; i++)
			{
				InnerTuioObject ito = (InnerTuioObject)iTuioObjects[i];
				if (ito.localSessionID == io.localSessionID){
					continue; // should not happen ....
				}
				args.Add((int)ito.localSessionID);
			}
			if (mode == MODE_OBJECT){
				foreach (InnerTuioToken it in tokens){
					if (it.localSessionID == io.localSessionID){
						continue; // should not happen ....
					}
					args.Add((int)it.localSessionID);
				}
			}
			args.Add((int)io.localSessionID);
			SendMessage("/tuio/2Dobj", args);
			SendFrameID("/tuio/2Dobj");

			args.Clear();
			args.Add("alive");
			for (int i = 0; i < iTuioObjects.Count; i++)
			{
				InnerTuioObject ito = (InnerTuioObject)iTuioObjects[i];
				if (ito.localSessionID == io.localSessionID){
					continue; // should not happen ....
				}
				args.Add((int)ito.localSessionID);
			}
			if (mode == MODE_OBJECT){
				foreach (InnerTuioToken it in tokens){
					if (it.localSessionID == io.localSessionID){
						continue; // should not happen ....
					}
					args.Add((int)it.localSessionID);
				}
			}
			SendMessage("/tuio/2Dobj", args);
			SendFrameID("/tuio/2Dobj");
		}

		public void addTuioObject(TuioObject o)
		{
			if (debugAddRemove) {
				Console.WriteLine("Add Object " + o.SymbolID + " (" + o.SessionID + ") " + o.X + " " + o.Y + " " + o.Angle);
			}
			//fwd
			lastLocalSessionID++;
			InnerTuioObject io = new InnerTuioObject(o, lastLocalSessionID);
			iTuioObjects.Add(io);
			SetTuioObject(io, true, false);
		}

		public void updateTuioObject(TuioObject o)
		{
			if (debugObjectUpdate) {
				Console.WriteLine("Update Object " + o.SymbolID + " " + o.SessionID + " " + o.X + " " + o.Y + " " + o.Angle + " " + o.MotionSpeed + " " + o.RotationSpeed + " " + o.MotionAccel + " " + o.RotationAccel);
			}
			// fwd
			InnerTuioObject ioToUpdate = null;
			foreach(InnerTuioObject io in iTuioObjects){
				if (io.obj.SessionID == o.SessionID){
					ioToUpdate = io;
					break;
				}
			}
			if (ioToUpdate != null){
				ioToUpdate.obj = o;
				SetTuioObject(ioToUpdate, false, false);
			}
		}

		public void removeTuioObject(TuioObject o)
		{
			if (debugAddRemove) {
				Console.WriteLine("Remove Object " + o.SymbolID + " (" + o.SessionID + ")");
			}
			//  fwd
			InnerTuioObject ioToRemove = null;
			foreach(InnerTuioObject io in iTuioObjects){
				if (io.obj.SessionID == o.SessionID){
					ioToRemove = io;
					break;
				}
			}
			if (ioToRemove != null){
				ioToRemove.obj = o;
				iTuioObjects.Remove(ioToRemove);
				SendRemoveTuioObject(ioToRemove, false);
			}
		}
		#endregion


		// ----------------------------------------------------------------------
		#region TUIO Blob
		private void SetTuioBlob(InnerTuioBlob ib, bool add, bool isToken){
			if (isToken && mode != MODE_OBJECT) { return; }
			TuioBlob b = ib.blob;
			TuioBlob nb = new TuioBlob(b.TuioTime, ib.localSessionID, b.BlobID, b.X, b.Y, b.Angle, b.Width, b.Height, b.Area);
			nb.update(b);
			for (int i = 0; i < listenerList.Count; i++)
			{
				TuioListener listener = (TuioListener)listenerList[i];
				if (add){
					if (listener != null) listener.addTuioBlob(nb);
				}
				else {
					if (listener != null) listener.updateTuioBlob(nb);
				}
			}
			if (!isToken){
				for (int i = 0; i < wallTokenListenerList.Count; i++)
				{
					TuioWallTokenListener listener = (TuioWallTokenListener)wallTokenListenerList[i];
					if (add){
						if (listener != null) listener.addTuioBlob(nb);
					}
					else {
						if (listener != null) {
							listener.updateTuioBlob(nb);
						}
					}
				}
			}
			if (sender == null) {
				return;
			}
			List<object> args = new List<object>();
			args.Add("set");
			args.Add((int)ib.localSessionID);
			args.Add((int)ib.blob.BlobID);
			args.Add((float)ib.blob.X);
			args.Add((float)ib.blob.Y);
			args.Add((float)ib.blob.Angle);
			args.Add((float)ib.blob.Width);
			args.Add((float)ib.blob.Height);
			args.Add((float)ib.blob.Area);
			args.Add((float)ib.blob.XSpeed);
			args.Add((float)ib.blob.YSpeed);
			args.Add((float)ib.blob.RotationSpeed);
			args.Add((float)ib.blob.MotionAccel);
			args.Add((float)ib.blob.RotationAccel);

			SendMessage("/tuio/2Dblb", args);
			SendFrameID("/tuio/2Dblb");
		}

		private void SendRemoveTuioBlob(InnerTuioBlob ib, bool isToken) {
			if (isToken && mode != MODE_OBJECT) { return; }
			TuioBlob b = ib.blob;
			TuioBlob nb = new TuioBlob(b.TuioTime, ib.localSessionID, b.BlobID, b.X, b.Y, b.Angle, b.Width, b.Height, b.Area);
			nb.update(b);
			for (int i = 0; i < listenerList.Count; i++)
			{
				TuioListener listener = (TuioListener)listenerList[i];
				if (listener != null) listener.removeTuioBlob(nb);
			}
			if (!isToken){
				for (int i = 0; i < wallTokenListenerList.Count; i++)
				{
					TuioWallTokenListener listener = (TuioWallTokenListener)wallTokenListenerList[i];
					if (listener != null) listener.removeTuioBlob(nb);
				}
			}
			if (sender == null) {
				return;
			}
			List<object> args = new List<object>();
			args.Add("alive");
			for (int i = 0; i < iTuioBlobs.Count; i++)
			{
				InnerTuioBlob itb = (InnerTuioBlob)iTuioBlobs[i];
				if (itb.localSessionID == ib.localSessionID){
					continue; // should not happen ....
				}
				args.Add((int)itb.localSessionID);
			}
			if (mode == MODE_OBJECT){
				foreach (InnerTuioToken it in tokens){
					if (it.localSessionID == ib.localSessionID){
						continue; // should not happen ....
					}
					args.Add((int)it.localSessionID);
				}
			}
			args.Add((int)ib.localSessionID);
			SendMessage("/tuio/2Dblb", args);
			SendFrameID("/tuio/2Dblb");

			args.Clear();
			args.Add("alive");
			for (int i = 0; i < iTuioBlobs.Count; i++)
			{
				InnerTuioBlob itb = (InnerTuioBlob)iTuioBlobs[i];
				if (itb.localSessionID == ib.localSessionID){
					continue; // should not happen ....
				}
				args.Add((int)itb.localSessionID);
			}
			if (mode == MODE_OBJECT){
				foreach (InnerTuioToken it in tokens){
					if (it.localSessionID == ib.localSessionID){
						continue; // should not happen ....
					}
					args.Add((int)it.localSessionID);
				}
			}
			SendMessage("/tuio/2Dblb", args);
			SendFrameID("/tuio/2Dblb");
		}

		public void addTuioBlob(TuioBlob b)
		{
			if (debugAddRemove) {
				Console.WriteLine("Add Blob " + b.BlobID + " (" + b.SessionID + ") " + b.X + " " + b.Y + " " + b.Angle + " " + b.Width + " " + b.Height + " " + b.Area);
			}
			//  fwd
			lastLocalSessionID++;
			InnerTuioBlob ib = new InnerTuioBlob(b, lastLocalSessionID);
			iTuioBlobs.Add(ib);
			SetTuioBlob(ib, true, false);
		}

		public void updateTuioBlob(TuioBlob b)
		{
			if (debugBlobUpdate) {
				Console.WriteLine("Update Blob " + b.BlobID + " (" + b.SessionID + ") " + b.X + " " + b.Y + " " + b.Angle + " " + b.Width + " " + b.Height + " " + b.Area + " " + b.MotionSpeed + " " + b.RotationSpeed + " " + b.MotionAccel + " " + b.RotationAccel);
			}
			// fwd
			InnerTuioBlob ibToUpdate = null;
			foreach(InnerTuioBlob ib in iTuioBlobs){
				if (ib.blob.SessionID == b.SessionID){
					ibToUpdate = ib;
					break;
				}
			}
			if (ibToUpdate != null){
				ibToUpdate.blob = b;
				SetTuioBlob(ibToUpdate, false, false);
			}
		}

		public void removeTuioBlob(TuioBlob b)
		{
			if (debugAddRemove) {
				Console.WriteLine("Remove Blob " + b.BlobID + " (" + b.SessionID + ")");
			}
			//  fwd
			InnerTuioBlob ibToRemove = null;
			foreach(InnerTuioBlob ib in iTuioBlobs){
				if (ib.blob.SessionID == b.SessionID){
					ibToRemove = ib;
					break;
				}
			}
			if (ibToRemove != null){
				ibToRemove.blob = b;
				iTuioBlobs.Remove(ibToRemove);
				SendRemoveTuioBlob(ibToRemove, false);
			}
		}
		#endregion

		// -----------------------------------------------------------------------------
		#region refresh
		private static void GetSubsets(List<int> superSet, int k, int idx, HashSet<int> current, List<HashSet<int>> solution) {
	    	//successful stop clause
    		if (current.Count == k) {
        		solution.Add(new HashSet<int>(current));
        		return;
    		}
    		//unseccessful stop clause
    		if (idx == superSet.Count) return;
    		int x = superSet[idx];
    		current.Add(x);
    		//"guess" x is in the subset
    		GetSubsets(superSet, k, idx+1, current, solution);
    		current.Remove(x);
    		//"guess" x is not in the subset
    		GetSubsets(superSet, k, idx+1, current, solution);
		}

		private static List<HashSet<int>> GetSubsets(List<int> superSet, int k) {
    		List<HashSet<int>> res = new List<HashSet<int>>();
    		GetSubsets(superSet, k, 0, new HashSet<int>(), res);
    		return res;
		}

		public void refresh(TuioTime frameTime)
		{
			// TODO: do the stuff at refresh only :) done for tokens	
			if (logWriter != null){
				logWriter.WriteLine("REFRESH " + logwatch.ElapsedMilliseconds + " " + frameTime.TotalMilliseconds);
			}

			for(int i = 0; i < tokens.Count; i++){
				InnerTuioToken it = tokens[i];
				if (it.dirty){
					it.dirty = false;
					if (debugTokenUpdate){
						Console.WriteLine("Refresh Update Token: " + it.twk.TokenID + " sid: "+ it.twk.SessionID + " sym:"+ it.twk.SymbolID +" x: "+ it.twk.X +" y: "+ it.twk.Y+" a: "+ it.twk.Angle*180/Math.PI);
					}
					SetTuioToken(it, false);
					InnerTuioObject tmpo = new InnerTuioObject(it.obj, it.localSessionID);
					SetTuioObject(tmpo, false, true);
					InnerTuioBlob tmpb = new InnerTuioBlob(it.blob, it.localSessionID);
					SetTuioBlob(tmpb, false, true);
					InnerTuioCursor tmpc = new InnerTuioCursor(it.cur, it.localSessionID);
					SetTuioCursor(tmpc, true);
				}
			}

			// free the waitCursors
			List<InnerTuioCursor> tmpics = new List <InnerTuioCursor>();
			foreach(InnerTuioCursor ic in waitCursors){
				if (frameTime.TotalMilliseconds - ic.downTime > waitCursorsDelay){
					tmpics.Add(ic);
				}
			}
			foreach(InnerTuioCursor ic in tmpics){
				ReallyAddTuioCursor(ic);
				waitCursors.Remove(ic);
			}

			// recognize token !
			bool restart = true;
			while(restart){
				restart = false;
				if (waitCursors.Count >= 3){
					List<int> inList = new List<int>();
					for (int i = 0; i < waitCursors.Count; i++) { inList.Add(i); }
					List<HashSet<int>> subsets = GetSubsets(inList, 3);
					// Console.WriteLine("Recognize Num subset: " + subsets.Count);
					if (logWriter != null) {
						logWriter.WriteLine("Recognize; Num cusors: " +  waitCursors.Count + ", Num subset: " + subsets.Count);
						int ii = 1;
						foreach(HashSet<int> ss in subsets){
							logWriter.Write("\t Set " + ii + ": "); ii++;
							foreach(int  j in ss) {
								logWriter.Write("\t" + j);
							}
							logWriter.Write("\n");
						}
					}
					foreach(HashSet<int> ss in subsets){
						List<TouchPoint> tmp = new List<TouchPoint>();
						foreach(int  j in ss) {
							tmp.Add(waitCursors[j].tp);
						}
						TokenTemplate t = recognizer.recognize(tmp);
						if (t != null && t.Distance <= minSquaredDistTokenPairsTolerance) {
							int sym = START_TOKEN_SYMB;
							foreach(TokenTemplate tplt in recognizer.tokenTemplates){
								if (tplt.tokenID == t.tokenID){
									break;
								}
								sym++;
							}
							lastLocalSessionID++;
							InnerTuioToken newtoken = new InnerTuioToken(lastLocalSessionID, sym, t, waitCursors); 
							tokens.Add(newtoken);
							if (debugAddRemove){
								Console.WriteLine("Add Token: " + newtoken.template.tokenID + " sid: "+ newtoken.localSessionID + " sym:"+ newtoken.obj.SymbolID +" x: "+ newtoken.obj.X +" y: "+ newtoken.obj.Y + " a: "+
									newtoken.obj.Angle*180/Math.PI);
							}
							if (logWriter != null){
								logWriter.WriteLine("Add Token: " + newtoken.template.tokenID + " sid: "+ newtoken.localSessionID + " sym:"+ newtoken.obj.SymbolID +" dist:"+ t.Distance + " x: "+ newtoken.obj.X +" y: "+ newtoken.obj.Y +
								" sw: "+ screenWidthMM + " sh: "+ screenHeightMM);
								int ii = 0;
								foreach (TouchPoint tp in tmp){
									ii++;
									logWriter.WriteLine("P"+ii+": " + tp.x + " " + tp.y);
								}
							}
							SetTuioToken(newtoken, true);
							InnerTuioObject tmpo = new InnerTuioObject(newtoken.obj, newtoken.localSessionID);
							SetTuioObject(tmpo, true, true);
							InnerTuioBlob tmpb = new InnerTuioBlob(newtoken.blob, newtoken.localSessionID);
							SetTuioBlob(tmpb, true, true);
							InnerTuioCursor tmpc = new InnerTuioCursor(newtoken.cur, newtoken.localSessionID);
							ReallyAddTuioCursor(tmpc, true);

							tmpics = new List <InnerTuioCursor>();
							foreach(int  j in ss) {
								tmpics.Add(waitCursors[j]);
							}
							foreach(InnerTuioCursor ic in tmpics){
								waitCursors.Remove(ic);
							}
							
							//Console.WriteLine("wait cursor cleaned: " + waitCursors.Count);
							restart = true;
							break;
						}
						else{
							// No token, clean ...
							// Console.WriteLine("\t no token");
							if (logWriter != null){
								logWriter.WriteLine("Failed Add Token: ");
								int ii = 0;
								foreach (TouchPoint tp in tmp){
									ii++;
									logWriter.WriteLine("P"+ii+": " + tp.x + " " + tp.y);
								}
							}
						}
					} // foreach subset
				}
				//Console.WriteLine("Recognize END pre High LOOP "+ restart);
			} // while restart

			// ----
			for (int i = 0; i < listenerList.Count; i++)
			{
				TuioListener listener = (TuioListener)listenerList[i];
				if (listener != null) listener.refresh(frameTime);
			}
			for (int i = 0; i < wallTokenListenerList.Count; i++)
			{
				TuioWallTokenListener listener = (TuioWallTokenListener)wallTokenListenerList[i];
				if (listener != null) listener.refresh(frameTime);
			}
			if (sender == null){
				return;
			}
			// todo sender ... nothing fseq send when needed ????
			// yes, but maybe we can send less fseq

		}
#endregion

	} // class WallTokensTuio : TuioListener
} // namespace